/* src/vm/jit/arm/asmpart.S - Java-C interface functions for arm

   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
   Institut f. Computersprachen - TU Wien

   This file is part of CACAO.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   Contact: cacao@complang.tuwien.ac.at

   Authors: Michael Starzinger

   $Id: asmpart.S,v 1.9 2005/07/12 14:39:37 michi Exp $

*/


#include "config.h"
#include "vm/jit/arm/offsets.h"
#include "vm/jit/arm/asmoffsets.h"
#include "vm/jit/arm/md-asm.h"


	.file "asmpart.S"
	.text
	.align 2


/********************* exported functions and variables ***********************/

	.globl asm_calljavafunction
	.globl asm_calljavafunction_int

	.globl asm_calljavafunction2
	.globl asm_calljavafunction2int
	.globl asm_calljavafunction2long
	.globl asm_calljavafunction2float
	.globl asm_calljavafunction2double

	.globl asm_call_jit_compiler

	.globl asm_handle_exception

	.globl asm_wrapper_patcher

	.globl asm_builtin_arraycheckcast
	.globl asm_builtin_aastore
	.globl asm_builtin_idiv
	.globl asm_builtin_irem
	.globl asm_builtin_ldiv
	.globl asm_builtin_lrem


	/* TODO implement these */
	.globl asm_sync_instruction_cache

	.globl asm_handle_builtin_exception
	.globl asm_handle_nat_exception

	.globl asm_perform_threadswitch
	.globl asm_initialize_thread_stack
	.globl asm_switchstackandcall
	.globl asm_getcallingmethod
	.globl asm_criticalsections
	.globl asm_getclassvalues_atomic

	.globl asm_throw_and_handle_exception
	.globl asm_throw_and_handle_hardware_arithmetic_exception

	.globl asm_prepare_native_stackinfo
	.globl asm_remove_native_stackinfo


/********************* function asm_calljavafunction ***************************
*                                                                              *
*   This function calls a Java-method (which possibly needs compilation)       *
*   with up to 4 address parameters.                                           *
*                                                                              *
*   This functions calls the JIT-compiler which eventually translates the      *
*   method into machine code.                                                  *
*                                                                              *
*   C-prototype:                                                               *
*    javaobject_header *asm_calljavamethod (methodinfo *m,                     *
*         void *arg1, void *arg2, void *arg3, void *arg4);                     *
*                                                                              *
*******************************************************************************/

	/*.ascii  "calljavafunction\0\0"*/

	/*.ascii  "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"*/

asm_calljavafunction:
asm_calljavafunction_int:
	SAVE_SCRATCH_REGISTERS()          /* save our personal scratch regs   */
	add   fp, sp, #4*4                /* FP points to our arguments       */

	sub   sp, sp, #4                  /* this stackframe will be dropped  */
	str   a1, [sp]                    /* save method pointer to stack     */

	mov   a1, a2                      /* reorder remaining arguments      */
	mov   a2, a3
	mov   a3, a4
	ldr   a4, [fp]                    /* arg4 is passed to us by stack    */

        stmfd sp!, {r0 - r3}              /* call the debugger                */
        mov   r3, #0
        add   r2, sp, #4
        mov   r1, r0 
        mov   r0, #0xea
        @ bl    asm_debug_intern
        ldmfd sp!, {r0 - r3}

	bl    asm_call_jit_compiler       /* call the JIT-compiler            */

	stmfd sp!, {r0 - r3}              /* call the debugger                */
	mov   r3, #0
	mov   r2, sp
	mov   r1, r0
	mov   r0, #0xeb
	@ bl    asm_debug_intern
	ldmfd sp!, {r0 - r3}

	RESTORE_SCRATCH_REGS_AND_RETURN() /* return to caller, restore regs   */

asm_calljava_xhandler:
	mov   a1, xptr                    /* exception pointer is arg1        */
	bl    builtin_throw_exception     /* throw the exception              */
	mov   res1, #0                    /* return NULL                      */
	RESTORE_SCRATCH_REGS_AND_RETURN() /* return to caller, restore regs   */

asm_calljavafunction2:
asm_calljavafunction2int:
asm_calljavafunction2long:
asm_calljavafunction2float:
asm_calljavafunction2double:
	SAVE_SCRATCH_REGISTERS()          /* save our personal scratch regs   */

	mov   fp, a1                      /* save method pointer for compiler */
	mov   v6, a4                      /* pointer to arg block             */
	mov   v7, a2                      /* arg count                        */

	cmp    v7, #0                      /* reorder arguments!               */
	ldrgt  a1, [v6, #offjniitem]
	subgts v7, v7, #1
	ldrgt  a2, [v6, #(offjniitem+sizejniblock*1)]
	subgts v7, v7, #1
	ldrgt  a3, [v6, #(offjniitem+sizejniblock*2)]
	subgts v7, v7, #1
	ldrgt  a4, [v6, #(offjniitem+sizejniblock*3)]
	subgts v7, v7, #1

	/* TODO: reorder remaining stack arguments! */
	movgt a1, #0x99 /* only for debug */
	bgt   asm_debug /* remove it */

	sub   sp, sp, #4                  /* this stackframe will be dropped  */
	str   fp, [sp]                    /* save method pointer to stack     */

	stmfd sp!, {r0 - r3}              /* call the debugger                */
	mov   r3, #0
	add   r2, sp, #4
	mov   r1, r0
	mov   r0, #0xea
	@ bl    asm_debug_intern
	ldmfd sp!, {r0 - r3}

	bl    asm_call_jit_compiler       /* call the JIT-compiler            */

	stmfd sp!, {r0 - r3}              /* call the debugger                */
	mov   r3, #0
	mov   r2, sp
	mov   r1, r0
	mov   r0, #0xeb
	@ bl    asm_debug_intern
	ldmfd sp!, {r0 - r3}

	RESTORE_SCRATCH_REGS_AND_RETURN() /* return to caller, restore regs   */


/****************** function asm_call_jit_compiler *****************************
*                                                                              *
*   Invokes the compiler for untranslated JavaVM methods.                      *
*   What this method does:
*    - save args and LR
*    - fire up jit_compile (pass methodinfo pointer)
*    - restore args and LR
*    - drop stackframe with methodinfo pointer
*    - check for exceptions
*    - eventually write back new method pointer
*    - call jit code (wich will then return to caller)
*                                                                              *
*   ATTENTION - ATTENTION:                                                     *
*   The Stack contains a pointer to the method info structure (prepared        *
*   by createcompilerstub or other caller). This Stackframe (4 Bytes) is       *
*   automatically dropped by this function.                                    *
*                                                                              *
*   Using the return address in R26 and the            *
*   offset in the LDA instruction or using the value in methodptr R28 the      *
*   patching address for storing the method address can be computed:           *
*                                                                              *
*   method address was either loaded using                                     *
*   M_LDQ (REG_PV, REG_PV, a)        ; invokestatic/special    ($27)           *
*   M_LDA (REG_PV, REG_RA, low)                                                *
*   M_LDAH(REG_PV, REG_RA, high)     ; optional                                *
*   or                                                                         *
*   M_LDQ (REG_PV, REG_METHODPTR, m) ; invokevirtual/interface ($28)           *
*   in the static case the method pointer can be computed using the            *
*   return address and the lda function following the jmp instruction          *
*                                                                              *
*******************************************************************************/

asm_call_jit_compiler:
	ldr   itmp1, [lr, #-4]            /* load the instruction that jumped */
	and   itmp3, itmp1, #0x0ff00000   /* was it a MOV PC, IP ???          */
	teq   itmp3, #0x01a00000
	movne itmp3, #0                   /* no -> no writeback!              */
	bne   asm_call_jit_compiler_do
	ldr   itmp1, [lr, #-12]           /* yes -> load LDR IP, ___          */
	mov   itmp3, itmp1, lsl #20       /* try to isolate offset            */
	mov   itmp3, itmp3, lsr #20
	mov   itmp1, itmp1, lsr #16       /* try to isolate base register     */
	and   itmp1, itmp1, #0x0f
	teq   itmp1, #0x0b                /* basereg == METHODPTR ???         */
	addeq itmp3, itmp2, itmp3         /* methodptr == METHODPTR + offset  */
	beq   asm_call_jit_compiler_do

	ldr   itmp1, [lr, #0]             /* tricky: recompute ip here        */
	and   itmp2, itmp1, #0x0ff00000   /* is it a SUB IP, PC, #___ ???     */
	teq   itmp2, #0x02400000
	andeq itmp2, itmp1, #0xff         /* try to isolate immediate         */
	subeq itmp1, lr, itmp2            /* recompute IP here                */
	addeq itmp1, itmp1, #8            /* no PC-relative addressing here   */
	subeq itmp3, itmp1, itmp3         /* methodptr == IP - offset         */
	movne itmp3, #0 /*TODO: hack*/

asm_call_jit_compiler_do:
	SAVE_ARGUMENT_REGISTERS()         /* save our argument registers & LR */

	ldr   a1, [sp, #20]               /* get methodinfo pointer ...       */
	str   itmp3, [sp, #20]            /* (save calculated methodptr)      */
	bl    jit_compile                 /* ... and pass it to jit_compile   */
	mov   ip, res1                    /* got pointer to new jit-code      */

	RESTORE_ARGUMENT_REGISTERS()      /* load our argument registers & LR */

	ldr   itmp3, [sp]                 /* (restore calculated methodptr)   */
	add   sp, sp, #4                  /* drop methodinfo pointer (!!!)    */

	tst   ip,ip                       /* check for exeption               */
	beq   asm_call_jit_compiler_exception

        stmfd sp!, {r0 - r3}              /* call the debugger                */
        mov   r3, lr
        mov   r2, ip
        mov   r1, itmp3
        mov   r0, #0x88
        @ bl    asm_debug_intern
        ldmfd sp!, {r0 - r3}

	tst   itmp3, itmp3                /* we came from asm_calljavafunction (methodptr = 0) */
	strne ip, [itmp3]                 /* no -> save new pointer!          */

	mov   pc, ip                      /* so call our method, it will return to caller */

asm_call_jit_compiler_exception:
	mov   xptr, #0x12
	mov   xpc, #0x34
	b     asm_handle_exception


/********************* function asm_handle_exception ***************************
*                                                                              *
*   This function handles an exception. It does not use the usual calling      *
*   conventions. The exception pointer is passed in REG_ITMP1 and the          *
*   pc from the exception raising position is passed in REG_ITMP2. It searches *
*   the local exception table for a handler. If no one is found, it unwinds    *
*   stacks and continues searching the callers.                                *
*                                                                              *
*   void asm_handle_exception (exceptionptr, exceptionpc);                     *
*                                                                              *
*******************************************************************************/

asm_handle_exception:
	mov a1, #0xff
	mov a2, xptr
	mov a3, xpc
	b asm_debug


/********************* function asm_wrapper_patcher ****************************
*
*   TODO: document me
*
*   Stack layout when calling patcher function:
*    12 (16)   return address into JIT code (patch position)
*     8 (12)   pointer to virtual java_objectheader
*     4 ( 8)   machine code (which is patched back later)
*     0 ( 4)   unresolved class/method/field reference
*   [-4 ( 0)   patcher function pointer to call ]
*    -4 ( 0)   saved IP of caller (caller needs it!)
*
*******************************************************************************/

asm_wrapper_patcher:
	SAVE_ARGUMENT_REGISTERS();
	add   a1, sp, #(1+5)*4    /* pass sp, skip patcher function pointer   */
	ldr   itmp3, [sp, #(0+5)*4] /* get patcher function pointer           */
	str   ip, [sp, #(0+5)*4]  /* save IP to stack, some patchers need it  */
	mov   lr, pc
	mov   pc, itmp3           /* call the patcher function                */
	ldr   ip, [sp, #(0+5)*4]  /* restore IP from stack, caller needs it   */
	tst   res1, res1          /* check for an exception                   */
	RESTORE_ARGUMENT_REGISTERS();

	ldr   itmp3,[sp, #4*4]    /* get return address (into JIT code)       */
	add   sp, sp, #5*4        /* remove stack frame                       */

	movne pc, itmp3           /* jump to new patched code (if no except.) */

asm_wrapper_patcher_exception:
	mov   xpc, itmp3          /* return address into JIT code is xpc      */
#if defined(USE_THREADS) && defined(NATIVE_THREADS)
	/* TODO: threads not implemented yet */
	b     asm_debug
#else
	mov   res1, /*_exceptionptr*/#0
	add   res1, pc, #8
#endif
	ldr   xptr, [res1]        /* get the exception pointer                */
	/*mov   itmp3, #0
	str   itmp3, [res1]*/       /* clear the exception pointer              */ 
	b     asm_handle_exception/* we have the IP of the calling java func. */
	.int  0xff
	.int  0xfe
	.int  0xfd


/* dummy functions() */

.equ IMB, 0xf00000
.equ sys_cacheflush, 0x9f0002
asm_sync_instruction_cache:
	mov   a1, #0
	mov   a2, #0xf0000000
	mov   a3, #0
	swi   #sys_cacheflush /* clean and invalidate the entire cache!!! */
	mov   pc, lr
asm_perform_threadswitch:
        mov   a4, #0x41
        b     asm_handle_exception
asm_initialize_thread_stack:
        mov   a4, #0x42
        b     asm_handle_exception 
asm_switchstackandcall:
        mov   a4, #0x43
        b     asm_handle_exception 
asm_getcallingmethod:
        mov   a4, #0x44
        b     asm_handle_exception 
asm_criticalsections:
        mov   a4, #0x45
        b     asm_handle_exception 
asm_getclassvalues_atomic:
	stmfd sp!, {r4, r5, r6}
	ldr   r4,[a1,#offbaseval]
	ldr   r5,[a1,#offdiffval]
	ldr   r6,[a2,#offbaseval]
	str   r4,[a3,#offcast_super_baseval]
	str   r5,[a3,#offcast_super_diffval]
	str   r6,[a3,#offcast_sub_baseval]
	ldmfd sp!, {r4, r5, r6}
	mov   pc, lr
asm_prepare_native_stackinfo:
        mov   a4, #0x47
        b     asm_handle_exception 
asm_remove_native_stackinfo:
	mov   a4, #0x48
	b     asm_handle_exception

asm_handle_builtin_exception:
asm_handle_nat_exception:
asm_throw_and_handle_exception:
asm_throw_and_handle_hardware_arithmetic_exception:
	mov   a4, #0x50
	b     asm_handle_exception


/************************ function asm_builtin_[] ******************************
*                                                                              *
*   Does some simple checks and calls the corresponding builtin function.      *
*                                                                              *
*******************************************************************************/

asm_builtin_arraycheckcast:
	stmfd sp!, {a1, lr}             /* save object pointer and LR         */
	bl    builtin_arraycheckcast
	tst   res1, res1                /* check return value                 */
	ldmfd sp!, {res1, lr}           /* return object pointer as result    */

	beq   asm_builtin_array_throw  /* throw exception if (false) returned */
	mov   pc, lr                   /* return to caller                    */
asm_builtin_array_throw:
	mov   pc, lr /* TODO: this is a hack */
	mov   a4, #0x11
	b     asm_handle_exception /* TODO! */

asm_builtin_aastore:
	teq   a1, #0
	beq   asm_builtin_aastore_null  /* throw exception if null pointer    */
	ldr   itmp1, [a1,#offarraysize] /* load size                          */
	add   itmp2, a1, a2, lsl #2     /* add index*4 to arrayref            */
	cmp   a2, itmp1                 /* do a bound check                   */
	bgt   asm_builtin_aastore_bound /* throw exception if out of bounds   */

	mov   a2, a3                    /* pass object as second argument     */
	stmfd sp!, {a2, itmp2, lr}      /* store things we need               */
	bl    builtin_canstore          /* builtin_canstore(arrayref, object) */
	tst   res1, res1                /* check return value                 */
	ldmfd sp!, {a2, itmp2, lr}

	beq   asm_builtin_aastore_throw /* throw exception if (false) returned */
	str   a2, [itmp2,#offobjarrdata]/* store object pointer in array      */
	mov   pc, lr                    /* ... and return to caller           */
asm_builtin_aastore_null:
	mov   a4, #0x1a
	b     asm_handle_exception /* TODO! */
asm_builtin_aastore_bound:
	mov   a4, #0x1b
	b     asm_handle_exception /* TODO! */
asm_builtin_aastore_throw:
	mov   pc, lr /* TODO: this is a hack */
	mov   a4, #0x1c
	b     asm_handle_exception /* TODO! */

asm_builtin_ldiv:
	teq   a3, #0
	teqeq a4, #0
	bne   builtin_ldiv
/* fall through */
	mov   a1, #0x2c
	b     asm_handle_exception /* TODO! */

asm_builtin_lrem:
	teq   a3, #0
	teqeq a4, #0
	bne   builtin_lrem
/* fall through */
	mov   a1, #0x2d
	b     asm_handle_exception /* TODO! */

asm_builtin_idiv:
	teq   a2, #0
	bne   builtin_idiv
/* fall through */
	mov   a4, #0x2a
	b     asm_handle_exception /* TODO! */

asm_builtin_irem:
	teq   a2, #0
        bne   builtin_irem
/* fall through */
	mov   a4, #0x2b
	b     asm_handle_exception /* TODO! */


/*
 * These are local overrides for various environment variables in Emacs.
 * Please do not remove this and leave it at the end of the file, where
 * Emacs will automagically detect them.
 * ---------------------------------------------------------------------
 * Local variables:
 * mode: asm
 * indent-tabs-mode: t
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 */
