/* src/vm/jit/arm/md-asm.h - assembler defines for arm ABI

   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
   Institut f. Computersprachen - TU Wien

   This file is part of CACAO.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   Contact: cacao@complang.tuwien.ac.at

   Authors: Michael Starzinger

   $Id: md-asm.h,v 1.2 2005/07/09 01:32:49 michi Exp $

*/


#ifndef _MD_ASM_H
#define _MD_ASM_H


/* register defines ***********************************************************/

#define res1    r0  /* result registers         */
#define res2    r1

#define a1      r0  /* argument registers       */
#define a2      r1
#define a3      r2
#define a4      r3

#define v1      r4  /* variable registers       */
#define v2      r5
#define v3      r6
#define v4      r7
#define v5      r8
#define v6      r9
#define v7      r10
#define v8      r11

#define pc      r15 /* program counter          */
#define lr      r14 /* return address           */
#define sp      r13 /* stack pointer            */
#define ip      r12 /* something like pv        */
#define fp      r11 /* frame pointer (not used) */

#define itmp1   v7  /* temporary scratch regs   */
#define itmp2   v8
#define itmp3   v6  /* TODO: maybe correct me */

#define xptr  itmp1 /* exception registers      */
#define xpc   itmp2


/* save and restore macros ****************************************************/

#define SAVE_ARGUMENT_REGISTERS() \
	stmfd   sp!, {a1,a2,a3,a4,lr}
#define SAVE_SCRATCH_REGISTERS() \
	stmfd   sp!, {itmp1,itmp2,itmp3,lr}

#define RESTORE_ARGUMENT_REGISTERS() \
	ldmfd   sp!, {a1,a2,a3,a4,lr}
#define RESTORE_SCRATCH_REGS_AND_RETURN() \
	ldmfd   sp!, {itmp1,itmp2,itmp3,pc}

#endif /* _MD_ASM_H */

