/* src/vm/jit/arm/md.c - machine dependent arm functions

   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
   Institut f. Computersprachen - TU Wien

   This file is part of CACAO.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   Contact: cacao@complang.tuwien.ac.at

   Authors: Michael Starzinger

   $Id: md.c,v 1.3 2005/07/09 01:59:09 michi Exp $

*/


#include <ucontext.h>

#include "config.h"
#include "vm/jit/arm/md-abi.h"
#include "vm/jit/arm/types.h"
#include "vm/exceptions.h"
#include "vm/stringlocal.h"
#include "vm/jit/asmpart.h"


/* md_init *********************************************************************

   Do some machine dependent initialization.

*******************************************************************************/

void md_init(void)
{
	/* do nothing here */
}


/* signal_handler_sigsegv ******************************************************

   NullPointerException signal handler for hardware null pointer check.

*******************************************************************************/

void signal_handler_sigsegv(int sig, siginfo_t *siginfo, void *_p)
{
	ucontext_t *_uc;
	mcontext_t *_mc;
	u4          instr;
	ptrint      base;

	_uc = (ucontext_t*) _p;
	_mc = &_uc->uc_mcontext;

	/*instr = *((s4*) (_mc->gregs[REG_PC]));*/instr=0x0;
	/*base = _mc->gregs[(instr >> 16) & 0x0f];*/base=0xff;

	log_text("!!! Segmentation fault");

	if (base == 0) {
		_mc->gregs[REG_ITMP1_XPTR] = (ptrint) string_java_lang_NullPointerException;
		_mc->gregs[REG_ITMP2_XPC] = _mc->gregs[REG_PC];
		_mc->gregs[REG_PC] = (ptrint) asm_throw_and_handle_exception;
	} else {
		throw_cacao_exception_exit(string_java_lang_InternalError,
		   "faulting address: %p (pc=%p, instr=%x, base=%p)\n",
		   siginfo->si_addr, _mc->gregs[REG_PC], instr, base);
	}
}


/*
 * These are local overrides for various environment variables in Emacs.
 * Please do not remove this and leave it at the end of the file, where
 * Emacs will automagically detect them.
 * ---------------------------------------------------------------------
 * Local variables:
 * mode: c
 * indent-tabs-mode: t
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 */

